/* $USAGI: tcpconnect.c,v 1.8 2004/07/06 07:21:18 yoshfuji Exp $ */

/*
 * Copyright (C) 2001,2002,2003,2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <malloc.h>

#ifdef TCP_MD5SIG
# define ENABLE_MD5SIG
#endif

#ifdef ENABLE_MD5SIG
# if HAVE_LINUX_TCP_H
#  include <linux/tcp.h>
# endif
#include "base64.h"
#endif

#define	PACKET_STRING	"TestPacket"

int setup_ip6opt(struct cmsghdr *cmsg, int type, int optlen)
{
	unsigned char *buf, *p;
	if (optlen < 8 || optlen > 255 * 8 || optlen % 8) {
		fprintf(stderr, "invalid optlen: %d\n", optlen);
		return -1;
	}

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = type;
	cmsg->cmsg_len = CMSG_LEN(optlen);

	p = CMSG_DATA(cmsg);

	memset(p, 0, optlen);

	p[1] = (optlen >> 3) - 1;

	p += 2;
	optlen -= 2;

	while (optlen > 0) {
		int len = optlen;

		if (len == 1) {
			/* Pad1 */
			p[0] = 0;
		} else {
			/* PadN */
			if (len > 257)
				len = 257;
			p[0] = 1;
			p[1] = len - 2;
		}

		p += len;
		optlen -= len;
	}

	return 0;
}

int main(int argc, char **argv) {
	const char *argv0 = argv[0];
	char *host, *serv;
	struct addrinfo hints, *res;
	int gai;
	int s;
	int len;
	char *buf, *cmsgbuf;
	int chunk_size = 100;
	int pref_privacy = 0;
	int packet_size = 100;
	int c, usage = 0;
#ifdef ENABLE_MD5SIG
	char *key = NULL;
	int keylen = 0;
#endif

	while((c = getopt(argc, argv, 
				"c"
				"d:"
				"h:"
#if ENABLE_MD5SIG
				"k:"
#endif
				"p:"
				"s:"
			  )) != EOF){
		switch (c) {
#if ENABLE_MD5SIG
		case 'k':
		    {
			int len = strlen(optarg);
			if (len < 0)
				goto help;
			free(key);
			key = malloc(b64_space(len));
			if (!key) {
				perror("malloc");
				exit(1);
			}
			keylen = b64_decode(key, optarg, len);
			if (keylen < 0) {
				printf("%s: base64 decode error: %s\n", argv0, optarg);
				exit(1);
			}
			break;
		    }
#endif
		case 'p':
			pref_privacy = atoi(optarg);
			break;
		case 's':
			packet_size = atoi(optarg);
			break;
		default:
			usage = 1;
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (usage || argc < 2 || chunk_size <= 0 || packet_size <= 0) {
help:
		printf("%s "
			"[-c] "
			"[-d dstopt_size] "
			"[-h hbhopt_size] "
#if defined(MSG_MORE)
			"[-m chunk_size] "
#endif
			"[-p pref_privacy] "
			"[-s packet_size] "
			"host port\n", argv0);
		exit(1);
	}
	host = argv[0];
	serv = argv[1];

	chunk_size = packet_size;
	buf = malloc(chunk_size);
	if (!buf) {
		perror("malloc");
		exit(1);
	}
	memset(buf, 0, chunk_size);
	snprintf(buf, chunk_size, "%s", PACKET_STRING);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	gai = getaddrinfo(host, serv, &hints, &res);
	if (gai) {
		printf("getaddrinfo() for %s port %s: %s\n",
			host, serv, gai_strerror(gai));
		exit(1);
	}

	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (s < 0) {
		perror("socket");
		exit(1);
	}

	if (res->ai_family == PF_INET6 && pref_privacy) {
#ifdef IPV6_PRIVACY
		if (setsockopt(s, IPPROTO_IPV6, IPV6_PRIVACY, &pref_privacy, sizeof(pref_privacy)) < 0)
			perror("setsockopt(IPV6_PRIVACY)");
#else
		printf("IPV6_PRIVACY is not supported.\n");
#endif
	}	

#ifdef ENABLE_MD5SIG
	if (key) {
# ifdef TCP_MD5SIG
#  ifndef TCP_MD5SIG_KEY
		/* FreeBSD / NetBSD */
		int md5sig = 1;
#  else
		/* Linux */
		struct tcp_md5sig_key md5sig;
#  endif
# elif defined(TCP_RFC2385)
		struct tcp_rfc2385_cmd md5sig;
# endif
# if defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385)
		/*XXX*/
		char nhost[NI_MAXHOST];
		struct addrinfo *peer;
		getnameinfo(res->ai_addr, res->ai_addrlen, nhost, sizeof(nhost), NULL, 0, NI_NUMERICHOST);
		hints.ai_flags = AI_NUMERICHOST;
		getaddrinfo(nhost, NULL, &hints, &peer);

		memset(&md5sig, 0, sizeof(md5sig));
#  ifdef TCP_MD5SIG_KEY
		md5sig.tcpsig_op = TCP_MD5SIG_KEY;
		md5sig.tcpsig_key = key;
		md5sig.tcpsig_keylen = keylen;
		md5sig.tcpsig_peerlen = peer->ai_addrlen;
		md5sig.tcpsig_peer = peer->ai_addr;
#  else
		if (peer->ai_addr->sa_family != AF_INET) {
			perror("unsupported family for MD5 Signature");
			exit(1);
		}
		md5sig.command = TCP_RFC2385_ADD;
		md5sig.address = ((struct sockaddr_in *)res0->ai_addr)->sin_addr.s_addr;	/*XXX*/
		md5sig.keylen = keylen;
		md5sig.key = key;
#  endif	/* TCP_MD5SIG_KEY */
# endif	/* defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385) */
		if (setsockopt(s, IPPROTO_TCP, TCP_MD5SIG, &md5sig, sizeof(md5sig)) < 0)
			perror("setsockopt(TCP_MD5SIG)");
# if defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385)
		freeaddrinfo(peer);
# endif
	}
#endif

	if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
		perror("connect");
		close(s);
		exit(1);
	}

	while (packet_size > 0) {
		struct iovec iov;

		if (packet_size <= chunk_size) {
			chunk_size = packet_size;
		}

		len = write(s, buf, chunk_size);
		if (len < 0) {
			perror("sendmsg");
			close(s);
			exit(1);
		}
		printf ("len = %d, sent = %d\n", chunk_size, len);
		packet_size -= chunk_size;
	}

	close(s);
	exit(0);
}

